/**********************************************************
 **
 ** jal_expr.c : JAL expression parsing
 **
 ** Copyright (c) 2004-2005, Kyle A. York
 ** All rights reserved
 **
 ***********************************************************/
#include <string.h>
#include "../libutils/mem.h"
#include "../libcore/pf_expr.h"
#include "jal_proc.h"
#include "jal_tokn.h"
#include "jal_expr.h"

value_t jal_parse_value(pfile_t *pf)
{
  char       *name;
  const char *ptr;
  value_t     val;

  ptr = pf_token_get(pf, pf_token_current);

  /* first look for _name_get; failing that, look for name */
  name = MALLOC(6 + strlen(ptr));
  sprintf(name, "_%s_get", ptr);
  val = pfile_value_find(pf, pfile_log_none, name);
  if (val) {
    if (variable_def_type_pointer == value_type_get(val)) {
      value_dereference(val);
    }
  } else {
    val = pfile_value_find(pf, pfile_log_none, ptr);
    if (!val) {
      sprintf(name, "_%s_put", ptr);
      val = pfile_value_find(pf, pfile_log_none, name);
      if (val) {
        pfile_log(pf, pfile_log_err, 
            "Attempted read from an OUT volatile parameter (%s)", ptr);
        value_release(val);
        val = VALUE_NONE;
        pf_token_get(pf, pf_token_next);
      }
    }
  }
  FREE(name);
  if (val) {
    pf_token_get(pf, pf_token_next);
    if (variable_def_type_function == value_type_get(val)) {
      value_t tval;

      tval = jal_parse_call(pf, val, boolean_false);
      value_release(val);
      val = tval;
      if (!tval) {
        pfile_log(pf, pfile_log_err, "no return value");
      }
    } else if (value_is_array(val)) {
      if (pf_token_is(pf, pf_token_current, pfile_log_err, "(")) {
        value_t n;

        pf_token_get(pf, pf_token_next);
        n = jal_parse_expr(pf);
        if (n) {
          /* n *= sz */
          value_t tmp1;

          value_use_ct_bump(val, ctr_bump_incr);
          tmp1 = value_clone(val);
          value_dereference(tmp1);

          if ((value_sz_get(tmp1) > 1) && !value_is_const(tmp1)) {
            value_t sz;
            value_t tmp;

            sz = pfile_constant_get(pf, value_sz_get(tmp1), VARIABLE_DEF_NONE);
            tmp = VALUE_NONE;
            pfile_cmd_op_add(pf, operator_mul, &tmp, n, sz);
            value_release(sz);
            value_baseofs_set(tmp1, tmp);
            value_release(tmp);
          } else {
            value_baseofs_set(tmp1, n);
          }
          if (!value_is_const(n) && value_is_const(val)) {
            variable_flag_set(value_variable_get(val), VARIABLE_FLAG_LOOKUP);
          }
          value_release(n);
          value_release(val);
          val = tmp1;
        }
        if (pf_token_is(pf, pf_token_current, pfile_log_err, ")")) {
          pf_token_get(pf, pf_token_next);
        }
      }
    }
  }
  return val;
}

/*
 * NAME
 *   value_get
 *
 * DESCRIPTION
 *   get an expression's value
 *
 * PARAMETERS
 *   pf    : return from pfile_open()
 *   var   : [out] holds result
 *
 * RETURN
 *
 * NOTES
 */
static value_t value_get(pfile_t *pf)
{
  value_t     val;
  const char *ptr;
  operator_t  op;

  /* valid values: 
       constant ('%' or '$' or decimal #)
       variable name (alpha + alpha|digit|'_')
       '(' -- subfunction
       '!' -- logical NOT
       '-' -- negation (two's complement)
       '~' -- complement (one's complement) */

  ptr = pf_token_get(pf, pf_token_current);
  op = operator_null;
  switch (*ptr) {
    case '!':
      op = ('!' == ptr[1]) ? operator_logical : operator_cmpb;
      /* fall through */
    case '-':
      if (operator_null == op) {
        op = operator_neg;
      }
      pf_token_get(pf, pf_token_next); /* this one's done */
      val = value_get(pf);   /* var is the next value */
      /* create the expression:
       *   _t# = var op
       *   and return _t#
       */
      if (val) {
        value_t tmp;

        if ((operator_neg == op) && value_is_const(val)) {
          variable_def_t def;

          def = variable_def_alloc(0, variable_def_type_integer,
              VARIABLE_DEF_FLAG_SIGNED, value_sz_get(val));
          tmp = pfile_constant_get(pf, -value_const_get(val), def);
          value_release(val);
          val = tmp;
        } else {
          tmp = VALUE_NONE;
          pfile_cmd_op_add(pf, op, &tmp, val, 0);
          value_release(val);
          val = tmp;
        }
      }
      break;
    case '(':
      pf_token_get(pf, pf_token_next);
      val = jal_parse_expr(pf);
      /* prepend the existing result to the new result */
      if (pf_token_is(pf, pf_token_current, pfile_log_err, ")")) {
        pf_token_get(pf, pf_token_next);
      }
      break;
    default:
      if (('_' == *ptr) || ISALPHA(*ptr)) {
        val = jal_parse_value(pf);
      } else {
        val = jal_token_to_constant(pf, pfile_log_err);
        if (val) {
          pf_token_get(pf, pf_token_next);
        }
      }
      break;
  }
  return val;
}

/*
 * NAME
 *   operator_get
 *
 * DESCRIPTION
 *   get an operator
 *
 * PARAMETERS
 *   expr : expression
 *   pf   : returned by pfile_open()
 *
 * RETURN
 *   0      : no error
 *   EINVAL : no operator present
 *
 * NOTES
 */
static operator_t operator_get(pfile_t *pf)
{
  static const struct {
    const char *token;
    operator_t     op;
  } operator_token_map[] = {
    {"*",    operator_mul},
    {"/",    operator_div},
    {"%",    operator_mod},
    {"+",    operator_add},
    {"-",    operator_sub},
    {"<<",   operator_shift_left},
    {">>",   operator_shift_right},
    {"<",    operator_lt},
    {"<=",   operator_le},
    {"==",   operator_eq},
    {"!=",   operator_ne},
    {">=",   operator_ge},
    {">",    operator_gt},
    {"&",    operator_andb},
    {"|",    operator_orb},
    {"^",    operator_xorb},
    {"!!",   operator_logical}
  };
  size_t ii;
  operator_t op;

  for (ii = 0; 
       (ii < COUNT(operator_token_map))
        && !pf_token_is(pf, pf_token_current, pfile_log_none,
          operator_token_map[ii].token); 
        ii++)
    ;
  if (ii == COUNT(operator_token_map)) {
    op = operator_null;
  } else {
    op = operator_token_map[ii].op;
    pf_token_get(pf, pf_token_next);
  }
  return op;
}

/*
 * NAME
 *   expr_parse
 *
 * DESCRIPTION
 *   parse an expression
 *
 * PARAMETERS
 *   dst   : [out] holds result
 *   pf    : parse file
 *   lhvar : for assignments, this is the base left-hand variable
 *
 * RETURN
 *
 * NOTES
 */
value_t jal_parse_expr(pfile_t *pf)
{
  operator_t  op;
  expr_t     *exprstk;
  value_t     val;

  exprstk = 0;
  op      = operator_null;
  do {
    val = value_get(pf);
    op = operator_get(pf);
    if (val) {
      pfile_expr_push(pf, &exprstk, val, op);
      value_release(val);
    }
  } while (operator_null != op);
  if (exprstk) {
    val = expr_val_get(exprstk);
    value_lock(val);
    expr_list_free(&exprstk);
  }
  return val;
}

